{-# OPTIONS_GHC -XPatternGuards -fbang-patterns #-}
module Controller (controller) where

-- ---------------------------------------------------------------------------

import System.IO
import Text.Printf (printf)
import Data.Complex
import Data.Maybe (listToMaybe)
import Data.List (sortBy)
import Control.Monad (guard)

import Message
import Command
import Sim

-- ---------------------------------------------------------------------------
-- パラメーター

-- simple small と同じもの
defaultHiddenParams :: HiddenParams
defaultHiddenParams
    = HiddenParams
    { accel = 2
    , brake = 3
    , rotAccel = 120
    }

defaultHiddenState :: HiddenState
defaultHiddenState
    = HiddenState
    { rotSpeed = 0
    }

-- コントローラ側でかかる時間
-- 値は適当
controllerOverhead :: Time
controllerOverhead = 5

-- ローバー側でコマンドを処理するのにかかる時間
-- 値は適当
vehicleOverhead :: Time
vehicleOverhead = 10

-- ---------------------------------------------------------------------------
-- メイン

controller :: Handle -> [Msg] -> IO ()
controller !h (Init iparams : ms) = do
  -- 何故か最初のコマンドは無視されることがあるのでダミーを送る
  sendCommand h (Nothing, Nothing)  
  loop h (iparams, defaultHiddenParams) ms

loop :: Handle -> Params -> [Msg] -> IO ()
loop h (ip, hp) = go hp defaultHiddenState []
  where
    go :: HiddenParams -> HiddenState -> [(Time, Command)] -> [Msg] -> IO ()
    go !hp !hs cs [] = undefined
    go !hp !hs cs (m : ms) =
        case m of 
          Telemetry{ timeStamp = t, vehicleState = vs } -> do
            let t' = t + 100
                (s'@(_,hs'), cs') = simulate (ip,hp) (vs, hs) cs t t'
            let offT = (75 + controllerOverhead + 75 + vehicleOverhead)
                fstCmdT = t + offT
                cmds = react (ip,hp) fstCmdT s'' cs'' (objects m)
                  where (s'', cs'') = simulate (ip,hp) s' cs' t' fstCmdT
                cs''' = cs' ++ zip [fstCmdT, (fstCmdT+vehicleOverhead)..] cmds
            sendCommands h cmds
            go hp hs' cs''' ms
          Event{ } ->
              go hp hs cs ms
          End{ timeStamp = tm, score = sc } -> do
              printf "score = %d\n" sc
              go hp defaultHiddenState cs ms

-- 状態から実行すべきコマンド列を計算
react :: Params -> Time -> State -> [(Time, Command)] -> [Object] -> [Command]
react !p !t s@(!vs, !hs) cmds objs = zip' as bs
  where
    pos = vehicleX vs :+ vehicleY vs
    speed = vehicleSpeed vs
    dir = vehicleDir vs * pi / 180
    (homeDistance, homeDir) = polar (-pos)
    (acceleration, turning) = vehicleCtl vs
    dirDiff = normalizeAngle (homeDir - dir)
    pos' = pos + mkPolar (speed / 10) dir
    homeDistance' = magnitude pos'
    mdobj = findMostDangerousObject vs objs
    as
      | Just (Object Crater (x,y,_)) <- mdobj
      , abs (phase ((x:+y) - pos) - dir * pi / 180) < pi / 9
      , magnitude ((x:+y) - pos) / speed < 500
        = [CmdB]
      | homeDistance < maxSpeed (fst p) * 2 && speed*2 > maxSpeed (fst p) = [CmdB]
      | homeDistance' > homeDistance = [CmdB]
      | speed < 100 = [CmdA]
      | otherwise = []
    bs
      | Just obj <- mdobj =
          case obj of
            Object k (x,y,_) ->
                if phase ((x:+y) - pos) - dir > 0
                then [CmdR]
                else [CmdL]
            Enemy (x, y, dir, speed) ->
                if phase ((x:+y) - pos) - dir > 0
                then [CmdR]
                else [CmdL]
      | dirDiff < -pi * 2 / 3 = [CmdR, CmdR]
      | dirDiff > pi * 2 / 3  = [CmdL, CmdL]
      | dirDiff < -pi / 4 = [CmdR]
      | dirDiff > pi / 4  = [CmdL]
      | dirDiff < -pi / 8 =
          case turning of
            HR -> [CmdL]
            R -> []
            _ -> [CmdR]
      | dirDiff > pi / 8 =
          case turning of
            HL -> [CmdR]
            L -> []
            _ -> [CmdL]
      | otherwise =
          case turning of
            HL -> [CmdR]
            L  -> [CmdR]
            HR -> [CmdL]
            R  -> [CmdL]
            ST -> []

zip' :: [a] -> [b] -> [(Maybe a, Maybe b)]
zip' [] [] = []
zip' (a:as) (b:bs) = (Just a, Just b) : zip' as bs
zip' as [] = [(Just a, Nothing) | a <- as]
zip' [] bs = [(Nothing, Just b) | b <- bs]

-- ---------------------------------------------------------------------------

-- [-pi, pi]
normalizeAngle :: Double -> Double
normalizeAngle x
    | x < -pi = normalizeAngle (x + 2*pi)
    | pi < x  = normalizeAngle (x - 2*pi)
    | otherwise = x

-- ---------------------------------------------------------------------------
-- 衝突判定

sqsum :: Double -> Double -> Double
sqsum x y = x*x + y*y

findMostDangerousObject :: VehicleState -> [Object] -> Maybe Object
findMostDangerousObject !vs objs = listToMaybe $ map fst $ sortBy c collidingObjs
  where
    c (_,d1) (_,d2) = compare d1 d2
    collidingObjs = do
      obj <- objs
      case obj of
        Object k (x, y, r) | k/=Homebase -> do
          guard $ collide a ((x,y),r,0,0) (if k==Crater then 2 else 1)
          return (obj, magnitude ((x:+y)-p) - (r + 5) - 0.5)
        Enemy (x, y, d, s) -> do
          guard $ collide a ((x,y),0.4,d*pi/180,s) 0.5
          return (obj, magnitude ((x:+y)-p) - (0.4 + 5) - 0.5)
        _ -> []
    a = ( (vehicleX vs, vehicleY vs)
        , 0.5
        , vehicleDir vs * pi / 180
        , vehicleSpeed vs
        )
    p = vehicleX vs :+ vehicleY vs

collide :: ((Double,Double), Double, Double, Double) ->
           ((Double,Double), Double, Double, Double) ->
           Double ->
           Bool
collide ((x1,y1),r1,th1,v1) ((x2,y2),r2,th2,v2) t = distance2 t' <= (r1+r2)^(2::Int)  where
  vx = v1 * cos th1 - v2 * cos th2
  vy = v1 * sin th1 - v2 * sin th2
  distance2 u = sqsum (x1+u*vx-x2) (y1+u*vy-y2)
  vv = sqsum vx vy
  t' = if (vv == 0) then 0 else max 0 (min t s)
  s = -((x1-x2)*vx+(y1-y2)*vy)/vv

-- ---------------------------------------------------------------------------
