{-# OPTIONS_GHC -cpp #-}
module Message where
import Data.List
import Debug.Trace

data Acceleration = Accelerating | Braking | Rolling deriving (Eq, Show)
instance Read Acceleration where
  readsPrec _ "a" = [(Accelerating,"")]
  readsPrec _ "b" = [(Braking,"")]
  readsPrec _ "-" = [(Rolling,"")]

data Turning = L | HL | R | HR | ST deriving (Eq, Show)
instance Read Turning where
  readsPrec _ "L" = [(HL,"")]
  readsPrec _ "l" = [(L,"")]
  readsPrec _ "-" = [(ST,"")]
  readsPrec _ "r" = [(R,"")]
  readsPrec _ "R" = [(HR,"")]

data ObjectKind = Boulder | Crater | Homebase | Martian deriving (Eq, Show)
instance Read ObjectKind where
  readsPrec _ "b" = [(Boulder,"")]
  readsPrec _ "c" = [(Crater,"")]
  readsPrec _ "h" = [(Homebase,"")]
  readsPrec _ "m" = [(Martian,"")]

-- Milliseconds
type Time = Int
-- Meters
type Length = Double
-- Meters per second
type Speed = Double
-- Degrees
type Dir = Double
-- Degrees per second
type AngularSpeed = Double

data Object
    = Object ObjectKind (Double, Double, Length) 
    | Enemy (Double, Double, Dir, Speed)
    deriving (Eq, Show)

instance Read Object where
  readsPrec _ s = [(readObj (words s), "")]

data VehicleState = VehicleState { vehicleCtl :: (Acceleration, Turning)
                                 , vehicleX :: !Double
                                 , vehicleY :: !Double
                                 , vehicleDir :: !Dir
                                 , vehicleSpeed :: !Speed
                                 } deriving (Eq, Show)

data InitParams
    = InitParams
    { dx :: !Double
    , dy :: !Double
    , timeLimit   :: !Time
    , minSensor   :: !Length
    , maxSensor   :: !Length
    , maxSpeed    :: !Speed
    , maxTurn     :: !AngularSpeed
    , maxHardTurn :: !AngularSpeed
    } deriving (Eq, Show)

data Msg 
    = Init !InitParams
    | Telemetry
    { timeStamp    :: !Time
    , vehicleState :: !VehicleState
    , objects :: [Object]
    }
    | Event
    { event :: !EventType
    , timeStamp :: !Time
    } 
    | End
    { timeStamp :: !Time
    , score :: !Int
    }
    deriving (Eq, Show)

data EventType
    = B -- crash
    | C -- fall into crater
    | K -- killed by Martian
    | S -- success
    deriving (Eq, Show, Read)

(...) :: Show b => (b -> c) -> (a -> b) -> (a -> c)

#ifdef DEBUG
(f ... g) x = let gx = g x in trace ("Response: "++show gx) (f gx)
#else
(...) = (.)
#endif

readMessages :: String -> [Msg]
readMessages = map (readMsg ... words) . separateBy ';'

separateBy :: Char -> String -> [String]
separateBy c = unfoldr phi
  where phi "" = Nothing
        phi s  = case break (c==) s of
                   (xs,_:ys) -> Just (xs,ys)
                   xsys      -> Just xsys

readMsg :: [String] -> Msg
readMsg ("I":dx:dy:tl:nsn:xsn:xsp:xtr:xhtr:[])
  = Init $ InitParams 
    { dx = read dx
    , dy = read dy
    , timeLimit   = read tl
    , minSensor   = read nsn
    , maxSensor   = read xsn
    , maxSpeed    = read xsp
    , maxTurn     = read xtr
    , maxHardTurn = read xhtr
    }

readMsg ("T":tst:vctl:vx:vy:vdir:vsp:objs)
  = Telemetry { timeStamp   = read tst
              , vehicleState = VehicleState { vehicleCtl  = readVCtl vctl
                                            , vehicleX     = read vx
                                            , vehicleY     = read vy
                                            , vehicleDir   = read vdir
                                            , vehicleSpeed = read vsp
                                            }
              , objects     = readObjs objs
              }
readMsg ("E":t:s:[])
  = End { timeStamp = read t
        , score     = read s
        }
readMsg ("B":t:[])
  = Event { event = B
          , timeStamp = read t
          }
readMsg ("C":t:[])
  = Event { event = C
          , timeStamp = read t
          }
readMsg ("K":t:[])
  = Event { event = K
          , timeStamp = read t
          }
readMsg ("S":t:[])
  = Event { event = S
          , timeStamp = read t
          }
readMsg unknown
  = error ("unknown message : "++ unwords unknown)

readVCtl :: String -> (Acceleration, Turning)
readVCtl [a,t] = (read [a], read [t])

readObj :: [String] -> Object
readObj (k:rs) = case read k of
  Martian -> case map read rs of
               [x,y,d,s] -> Enemy (x,y,d,s)
  k       -> case map read rs of
               [x,y,r]   -> Object k (x,y,r)

readObjs :: [String] -> [Object]
readObjs [] = []
readObjs (k:rs) = case read k of
  Martian -> case map read (take 4 rs) of
               [x,y,d,s] -> Enemy (x,y,d,s) : readObjs (drop 4 rs)
  k       -> case map read (take 3 rs) of
               [x,y,r]   -> Object k (x,y,r) : readObjs (drop 3 rs)
